@@ -15,29 +15,118 @@ import matplotlib.pyplot as plt
 ## Miscellaneous
 preset_colors = ['#000000e0', '#dbb243e0', '#2e42d3e0', '#e54fe3e0', '#f23434e0']
 
-## Cat Dataframe Format
-_cat_df_columns = ["x", "error", "y", "degfreed", "elower", "usd", "tag", "qnfmt", 'qnu1', 'qnu2', 'qnu3', 'qnu4', 'qnu5', 'qnu6', 'qnl1', 'qnl2', 'qnl3', 'qnl4', 'qnl5', 'qnl6', 'comment']
-_cat_df_dtypes = [np.float64, np.float64, np.float64, np.int16, np.float64, np.int16, np.int32, np.int16]+[np.int16]*12+[str]
-_cat_widths = [13, 8, 8, 2, 10, 3, 7, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 100]
+cat_dtypes = {
+  'x':			np.float,
+  'error':		np.float,
+  'y':			np.float,
+  'degfreed':	np.int,
+  'elower':		np.float,
+  'usd':		np.int,
+  'tag':		np.int,
+  'qnfmt':		np.int,
+  'qnu1':		np.int,
+  'qnu2':		np.int,
+  'qnu3':		np.int,
+  'qnu4':		np.int,
+  'qnu5':		np.int,
+  'qnu6':		np.int,
+  'qnl1':		np.int,
+  'qnl2':		np.int,
+  'qnl3':		np.int,
+  'qnl4':		np.int,
+  'qnl5':		np.int,
+  'qnl6':		np.int,
+}
 
-## Lin Dataframe Format
-_lin_df_columns = ['qnu1', 'qnu2', 'qnu3', 'qnu4', 'qnu5', 'qnu6', 'qnl1', 'qnl2', 'qnl3', 'qnl4', 'qnl5', 'qnl6', "x", "error", "weight", "comment"]
-_lin_df_dtypes = [np.int16]*12 + [np.float64, np.float64, np.float64]+[str]
-_lin_widths = range(0,37,3)
+lin_dtypes = {
+  'qnu1':		np.int,
+  'qnu2':		np.int,
+  'qnu3':		np.int,
+  'qnu4':		np.int,
+  'qnu5':		np.int,
+  'qnu6':		np.int,
+  'qnl1':		np.int,
+  'qnl2':		np.int,
+  'qnl3':		np.int,
+  'qnl4':		np.int,
+  'qnl5':		np.int,
+  'qnl6':		np.int,
+  'x':			np.float,
+  'error':		np.float,
+  'weight':		np.float,
+  'comment':	str,
+}
+
+cat_widths = {
+  'x':			13,
+  'error':		 8,
+  'y':			 8,
+  'degfreed':	 2,
+  'elower':		10,
+  'usd':		 3,
+  'tag':		 7,
+  'qnfmt':		 4,
+  'qnu1':		 2,
+  'qnu2':		 2,
+  'qnu3':		 2,
+  'qnu4':		 2,
+  'qnu5':		 2,
+  'qnu6':		 2,
+  'qnl1':		 2,
+  'qnl2':		 2,
+  'qnl3':		 2,
+  'qnl4':		 2,
+  'qnl5':		 2,
+  'qnl6':		 2,
+}
+
+qnlabels =  ['qnu1', 'qnu2', 'qnu3', 'qnu4', 'qnu5', 'qnu6', 'qnl1', 'qnl2', 'qnl3', 'qnl4', 'qnl5', 'qnl6']
 
 ## Egy Dataframe Format
-_egy_df_columns = ['iblk', 'indx', 'egy', 'err', 'pmix', 'we', ':', 'qn1', 'qn2', 'qn3', 'qn4', 'qn5', 'qn6']
-_egy_df_dtypes = [np.int16, np.int16, np.float64, np.float64, np.float64, np.int16, str]+[np.int16]*6
-_egy_df_widths = [0,6,11,29,47,58,63,64,67,70,73,76,79,82]
+egy_widths = {
+	'iblk':		 6,
+	'indx':		 5,
+	'egy':		18,
+	'err':		18,
+	'pmix':		11,
+	'we':		 5,
+	':':		 1,
+	'qn1':		 3,
+	'qn2':		 3,
+	'qn3':		 3,
+	'qn4':		 3,
+	'qn5':		 3,
+	'qn6':		 3,
+}
 
-## Helpers
+egy_dtypes = {
+	'iblk':		np.int,
+	'indx':		np.int,
+	'egy':		np.float,
+	'err':		np.float,
+	'pmix':		np.float,
+	'we':		np.int,
+	':':		np.int,
+	'qn1':		np.int,
+	'qn2':		np.int,
+	'qn3':		np.int,
+	'qn4':		np.int,
+	'qn5':		np.int,
+	'qn6':		np.int,
+}
+
+
+# Helper functions
 def str_to_stream(string):
 	return(io.StringIO(string))
 
 def column_to_numeric(val, force_int=False):
 	val = val.strip()
 	if val == "" or val == ":":
-		return(np.int16(-1))
+		if force_int:
+			return(np.iinfo(np.int).min)
+		else:
+			return(np.nan)
 	elif val[0].isalpha():
 		val = str(ord(val[0].upper())-55)+val[1:]
 
@@ -60,12 +149,31 @@ def rmdir(dir):
 		os.remove(os.path.join(dir, file))
 	os.rmdir(dir)
 
-## Format Functions
-def lin_to_df(fname):
-	dtypes = _lin_df_dtypes
-	widths = _lin_widths
-	column_names = _lin_df_columns
-	dtypes_dict = {column_names[i]:dtypes[i] for i in range(len(column_names))}
+def minmax(value, min=None, max=None):
+	if min and value < min:
+		return(min)
+	if max and value > max:
+		return(max)
+	return(value)
+
+# Format functions
+def cat_to_df(fname, sort=True):
+	widths = cat_widths.values()
+	columns = list(cat_dtypes.keys())
+	
+	converters = cat_dtypes.copy()
+	converters.update({key: lambda x: column_to_numeric(x, True) for key in columns[5:20]})
+	data = pd.read_fwf(fname, widths=widths, names=columns, converters=converters, skip_blank_lines=True).astype(cat_dtypes)
+	data["y"] = 10 ** data["y"]
+	data["filename"] = fname
+	
+	if sort:
+		data.sort_values("x", inplace=True)
+	return(data)
+
+def lin_to_df(fname, sort=True):
+	widths = range(0,37,3)
+	column_names = list(lin_dtypes.keys())
 	
 	data = []
 	tmp_file = open(fname, "r") if not isinstance(fname, io.StringIO) else fname
@@ -78,116 +186,81 @@ def lin_to_df(fname):
 			if len(tmp) == 2:
 				tmp.append("1.0000")
 
-			tmp_line_content =	  [column_to_numeric(line[i:j], True) for i, j in zip(_lin_widths[:-1], _lin_widths[1:])] + [column_to_numeric(x) for x in tmp[:3]]
+			tmp_line_content = [column_to_numeric(line[i:j], True) for i, j in zip(widths[:-1], widths[1:])] + [column_to_numeric(x) for x in tmp[:3]]
 			if len(tmp) == 4:
 				tmp_line_content.append(tmp[3].strip())
 			else:
 				tmp_line_content.append("")
 
 			data.append(tmp_line_content)
-	data = pd.DataFrame(data, dtype = np.float64)
-	data.columns = column_names
+	
+	data = pd.DataFrame(data, columns=column_names).astype(lin_dtypes)
 	data["filename"] = fname
-	data = data.astype(dtypes_dict)
 	
 	## Set correct columns for data
-	qn_labels = column_names[0:12]
-	QNs = len(qn_labels)
-	for i in range(len(qn_labels)):
-		tmp_unique = data[qn_labels[i]].unique()
-		if len(tmp_unique) == 1 and tmp_unique[0] == -1:
-			QNs = i
+	qns_labels = column_names[0:12]
+	noq = len(qns_labels)
+	for i in range(len(qns_labels)):
+		if all(np.iinfo(np.int).min == data[qns_labels[i]]):
+			noq = i
 			break
-	QNs = int(QNs/2)
-	columns_qn = [f"qnu{i+1}" for i in range(QNs)]+[f"qnl{i+1}" for i in range(QNs)]+[f"qnu{i+1}" for i in range(QNs, 6)]+[f"qnl{i+1}" for i in range(QNs, 6)]
-	new_column_names = columns_qn + column_names[12:] + ["filename"]
-	data.columns = new_column_names
+	noq = noq//2
 
-	data.sort_values("x", inplace = True)
-	return(data)
+	columns_qn = [f"qnu{i+1}" for i in range(noq)]+[f"qnl{i+1}" for i in range(noq)]+[f"qnu{i+1}" for i in range(noq, 6)]+[f"qnl{i+1}" for i in range(noq, 6)]
+	data.columns = columns_qn + list(data.columns[12:])
 
-def cat_to_df(fname):
-	dtypes = _cat_df_dtypes
-	widths = _cat_widths
-	column_names = _cat_df_columns
-	dtypes_dict = {column: dtype for column, dtype in zip(column_names, dtypes)}
-	converters = {key: column_to_numeric for key in column_names[5:20]}
-	for key in column_names[5:20]:
-		del dtypes_dict[key]
-	
-	data = pd.read_fwf(fname, widths=widths, names=column_names, converters=converters, skip_blank_lines=True, dtype=dtypes_dict)
-	data["filename"] = fname
-	data["y"] = 10 ** data["y"]
-	
-		
-	data.sort_values("x", inplace = True)
+	if sort:
+		data.sort_values("x", inplace=True)
 	return(data)
 
 def df_to_cat(df):
-	output = []
+	lines = []
 
 	for index, row in df.iterrows():
-		freq = row["x"]
-		if freq > 99999999.9999:
-			freq = 99999999.9999
-		elif freq < 0.0001:
-			freq = 0.0001
-		error = row["error"]
-		if error > 999.9999:
-			error = 999.9999
-		elif error < -99.9999:
-			error = -99.9999
+		freq = minmax(row["x"], 0.0001, 99999999.9999)
+		error = minmax(row["error"], -99.9999, 999.9999)
 		intens = np.log10(row["y"]) if row["y"] > 0 else 0
-		if intens < -99.999:
-			intens = -99.999
-		elif intens > 999.999:
-			intens = 999.999
+		intens = minmax(intens, -99.999, 999.9999)
 
-		QNs_string = ""
-		for QNlabel in ['qnu1', 'qnu2', 'qnu3', 'qnu4', 'qnu5', 'qnu6', 'qnl1', 'qnl2', 'qnl3', 'qnl4', 'qnl5', 'qnl6']:
-			if row[QNlabel] in ["", -1]:
-				QNs_string += "  "
+		qnsstring = ""
+		for qnlabel in qnlabels:
+			if row[qnlabel] == np.iinfo(np.int).min:
+				qnsstring += "  "
 			else:
-				QNs_string += f"{row[QNlabel]:2.0f}"
+				qnsstring += f"{row[qnlabel]:2.0f}"
 
-		output.append(f"{freq:13.4f}{error:8.4f}{intens:8.4f}{row['degfreed']:2.0f}{row['elower']:10.4f}{row['usd']:3.0f}{row['tag']:7.0f}{row['qnfmt']:4.0f}{QNs_string}{r'  // '+row['comment'] if row['comment'] not in ['', -1, np.NaN] else ''}")
+		lines.append(f"{freq:13.4f}{error:8.4f}{intens:8.4f}{row['degfreed']:2.0f}{row['elower']:10.4f}{row['usd']:3.0f}{row['tag']:7.0f}{row['qnfmt']:4.0f}{qnsstring}")
 	
-	output = "\n".join(output)
-	return(output)
+	return("\n".join(lines))
 
 def df_to_lin(df):
-	output = []
+	lines = []
+
 	for index, row in df.iterrows():
-		QNs_string = ""
-		pad_string = ""
-		for QNlabel in ['qnu1', 'qnu2', 'qnu3', 'qnu4', 'qnu5', 'qnu6', 'qnl1', 'qnl2', 'qnl3', 'qnl4', 'qnl5', 'qnl6']:
-			if row[QNlabel] in ["", -1]:
-				pad_string += "   "
+		qnsstring = ""
+		padstring = ""
+		for qnlabel in qnlabels:
+			if np.iinfo(np.int).min == row[qnlabel]:
+				padstring += "   "
 			else:
-				QNs_string += f"{row[QNlabel]:3.0f}"
-		QNs_string = QNs_string + pad_string
-		output.append(f"{QNs_string} {row['x']:13.4f} {row['error']:8.4f} {row['weight']:13.4f}{r'  // '+row['comment'] if row['comment'] not in ['', -1, np.NaN] else ''}")
+				qnsstring += f"{min(row[qnlabel], 999):3.0f}"
+		qnsstring = qnsstring + padstring
+		lines.append(f"{qnsstring} {min(row['x'], 99999999.9999):13.4f} {min(row['error'], 999.9999):8.4f} {min(row['weight'], 99999999.9999):13.4f}{r'  // '+row['comment'] if row['comment'] else ''}")
+		
+	return("\n".join(lines))
+
+def egy_to_df(fname, sort=True):
+	widths = egy_widths.values()
+	columns = list(egy_dtypes.keys())
+
+	converters = egy_dtypes.copy()
+	converters.update({key: lambda x: column_to_numeric(x, True) for key in column_names[-6:]})
+	data = pd.read_fwf(fname, widths=widths, names=columns, converters=converters, skip_blank_lines=True).astype(egy_dtypes)
 	
-	output = "\n".join(output)
-	return(output)
-
-def egy_to_df(fname):
-	dtypes = _egy_df_dtypes
-	egyindices = _egy_df_widths
-	column_names = _egy_df_columns
-	dtypes_dict = {column_names[i]:dtypes[i] for i in range(len(column_names))}
-
-	data = []
-	tmp_file = open(fname, "r") if not isinstance(fname, io.StringIO) else fname
-	with tmp_file as file:
-		for line in file:
-			if line.strip() == "" or line.startswith("#"):
-				continue
-			data.append([column_to_numeric(line[i:j]) for i,j in zip(egyindices[:-1], egyindices[1:])])
-	data = pd.DataFrame(data, dtype = np.float64)
-
-	data.columns = column_names
-	data = data.astype(dtypes_dict)
+	data["filename"] = fname
+	
+	if sort:
+		data.sort_values("egy", inplace=True)
 	return(data)
 
 def parvar_to_dict(fname):
@@ -415,7 +488,7 @@ def residuals(df_cat, df_lin, query_string=None, scatter_dict={}, head_number=20
 			for j in range(6):
 				qnu = row[f"qnu{j+1}"]
 				qnl = row[f"qnl{j+1}"]
-				if qnu != -1 and qnl != -1:
+				if qnu != np.iinfo(np.int).min and qnl != np.iinfo(np.int).min:
 					upper_state.append(str(int(qnu)))
 					lower_state.append(str(int(qnl)))
 			upper_state = ", ".join(upper_state)
